@@ -1,9 +1,4 @@
-/* This is a file used to develop and test interfacing routines for 
-   Lapack and Linpack. In particular the aim is to check that I can reproduce
-   R routines such as chol and La.svd by direct calls to Lapack and Linpack.
-
-   They have been tested against R routines, (but R's qr() routine was buggy, so only
-   unpivoted case was testable).
+/* Convenient C wrappers for calling LAPACK and LINPACK 
 
    See R-x/include/R_ext/Lapack.h...
 */
@@ -15,6 +10,88 @@
 #include <R_ext/Lapack.h>
 /*#include <dmalloc.h>*/
 
+void mgcv_mmult(double *A,double *B,double *C,int *bt,int *ct,int *r,int *c,int *n)
+/* Forms r by c product of B and C, transposing each according to bt and ct.
+   n is the common dimension of the two matrices, which are stored in R 
+   default column order form. Algorithm is inner loop optimized in each case 
+   (i.e. inner loops only update pointers by 1, rather than jumping).
+   
+  
+   r<-1000;c<-1000;n<-1000
+   A<-matrix(0,r,c);B<-matrix(rnorm(r*n),n,r);C<-matrix(rnorm(n*c),c,n)
+   system.time(
+   A<-matrix(.C("mgcv_mmult",as.double(A),as.double(B),as.double(C),as.integer(1),as.integer(1),
+          as.integer(r),as.integer(c),as.integer(n))[[1]],r,c))
+   range(as.numeric(t(B)%*%t(C)-A))
+   A<-matrix(0,r,c);B<-matrix(rnorm(r*n),n,r);C<-matrix(rnorm(n*c),n,c)
+   system.time(
+   A<-matrix(.C("mgcv_mmult",as.double(A),as.double(B),as.double(C),as.integer(1),as.integer(0),
+          as.integer(r),as.integer(c),as.integer(n))[[1]],r,c))
+   range(as.numeric(t(B)%*%C-A))
+   A<-matrix(0,r,c);B<-matrix(rnorm(r*n),r,n);C<-matrix(rnorm(n*c),c,n)
+   system.time(
+   A<-matrix(.C("mgcv_mmult",as.double(A),as.double(B),as.double(C),as.integer(0),as.integer(1),
+          as.integer(r),as.integer(c),as.integer(n))[[1]],r,c))
+   range(as.numeric(B%*%t(C)-A))
+   A<-matrix(0,r,c);B<-matrix(rnorm(r*n),r,n);C<-matrix(rnorm(n*c),n,c)
+   system.time(
+   A<-matrix(.C("mgcv_mmult",as.double(A),as.double(B),as.double(C),as.integer(0),as.integer(0),
+          as.integer(r),as.integer(c),as.integer(n))[[1]],r,c))
+   range(as.numeric(B%*%C-A))  
+*/
+
+{ double xx,*bp,*cp,*cp1,*cp2,*cp3,*ap,*ap1;
+  int br,cr,i,j;
+  if (*bt)
+  { if (*ct) /* A=B'C' */
+    { /* this one is really awkward: have to use first row of C' as working storage
+         for current A row; move most slowly through B */
+      for (i=0;i<*r;i++) /* update A *row-wise*, one full sweep of C per i */    
+      { cp1 = C + *c;
+        /* back up row 1 of C' (in A), and initialize current row of A (stored in row 1 of C') */
+        xx = *B;
+        for (ap=A,cp=C;cp<cp1;cp++,ap+= *r) { *ap = *cp; *cp *= xx;}  
+        B++;cp2=cp1;
+        for (j=1;j< *n;j++,B++) 
+	    for (xx= *B,cp=C;cp<cp1;cp++,cp2++) *cp += xx * *cp2;      
+        /* row i of A is now in row 1 of C', need to swap back */
+        for (ap=A,cp=C;cp<cp1;cp++,ap+= *r) { xx = *ap; *ap = *cp; *cp = xx;}
+        A++;
+      } 
+    } else /* A=B'C - easiest case: move most slowly through A*/
+    { br= *n;cr= *n;cp2 = C + *c * cr;
+      for (ap=A,cp1=C;cp1< cp2;cp1+=cr) for (bp=B,i=0;i< *r;i++,ap++)  
+      { for (xx=0.0,cp=cp1,cp3=cp1+ *n;cp< cp3;cp++,bp++) xx += *cp * *bp; /* B[k+br*i]*C[k+cr*j];*/
+        *ap=xx;
+      }
+    }
+  } else
+  { if (*ct) /* A=BC' - update A column-wise, move most slowly through C (but in big steps)*/
+    { cp = C;
+      for (j=0;j< *c;j++) /* go through columns of A, one full B sweep per j */
+      { ap1 = A + *r;C=cp;xx = *C;bp=B;
+        for (ap=A;ap<ap1;ap++,bp++) *ap = xx * *bp ;
+        C += *c;
+        for (i=1;i< *n;i++,C+= *c) 
+	for (xx=*C,ap=A;ap<ap1;ap++,bp++) *ap += xx * *bp;
+        A=ap1;cp++;
+      }
+    } else /* A=BC - update A column-wise, moving most slowly through C */
+    { for (j=0;j< *c;j++) /* go through columns of A, one full B sweep per j */
+      { ap1 = A + *r;xx = *C;bp=B;
+        for (ap=A;ap<ap1;ap++,bp++) *ap = xx * *bp ;
+        C++;
+        for (i=1;i< *n;i++,C++) 
+	for (xx=*C,ap=A;ap<ap1;ap++,bp++) *ap += xx * *bp;
+        A=ap1;
+      }
+    }
+  }
+} 
+
+
+
+
 void mgcv_chol(double *a,int *pivot,int *n,int *rank)
 /* a stored in column order, this routine finds the pivoted choleski decomposition of matrix a 
    library(mgcv)
@@ -94,7 +171,7 @@ void mgcv_svd(double *x,double *u,double *d,int *r,int *c)
 
 void mgcv_svd_full(double *x,double *vt,double *d,int *r,int *c)
 /* call LA_PACK svd routine to form x=UDV'. U returned in x. V' returned in vt.
-   assumed r > c. U is r by c. D is length c. V is c by c.
+   assumed r >= c. U is r by c. D is length c. V is c by c.
 
 # Here is R test code.....
 library(mgcv)
